/* WARNING! All changes made to this file will be lost! */

#ifndef W_SRC_SPTK_CONFIG_H_WAF
#define W_SRC_SPTK_CONFIG_H_WAF

#define SPTK_VERSION "3.8.9"
#define HAVE_FCNTL_H 1
#define HAVE_LIMITS_H 1
#define HAVE_STDLIB_H 1
#define HAVE_STRING_H 1

#endif /* W_SRC_SPTK_CONFIG_H_WAF */
